import json
import logging
import urllib.error
import urllib.parse
import urllib.request

LOG = logging.getLogger()

NEXTBUS_XML_FEED_URL = 'http://webservices.nextbus.com/service/publicXMLFeed'
NEXTBUS_JSON_FEED_URL = 'http://webservices.nextbus.com/service/publicJSONFeed'

class NextBusClient():
    """Minimalistic client for making requests using the NextBus API.

    This client makes no assumptions about the structure of the data returned by requests to
    NextBus, nor valid values for the query parameters in requests. All response content is returned
    as-is, and no validation is done on query parameters beyond which keys are required.

    All commands in revision 1.23 of the NextBus API are supported."""

    def __init__(self, output_format, agency=None, use_compression=True):
        """Arguments:
            output_format: (String) Indicates the format of the data returned by requests, either
                "json" or "xml".
            agency: (String) Name of a transit agency on NextBus. If an agency is specified for an
                instance of this class, an agency does not have to be provided with every request to
                the NextBus API (Other than the agencyList command).
            use_compression (Boolean) Indicates whether the response data from requests to NextBus
                should be compressed.
        """

        if not isinstance(output_format, str):
            raise TypeError('"output_format" must be a string.')

        if output_format.lower() not in ['json', 'xml']:
            raise ValueError('"output_format" must be either "json" or "xml".')

        if not isinstance(use_compression, bool):
            raise TypeError('"use_compression" must be a boolean.')

        self.output_format = output_format.lower()
        self.agency = agency
        self.use_compression = use_compression

    def get_agency_list(self):
        """Make a request to the NextBus API with the "agencyList" command to get the list of
        transit agencies with predictions on NextBus.

        Returns:
            If the output_format is "json": Dictionary containing the JSON returned by the request.
            If the output_format is "xml": String containing the XML returned by the request.

        Raises:
            urllib.error.HTTPError: If an HTTP error occurs when making the request to the NextBus
                API.
            json.decoder.JSONDecodeError: If the output_format is "json" and the response was not
                valid JSON.
        """

        params = {
            'command': 'agencyList'
        }

        return self._perform_request(params=params)

    def get_messages(self, route_tags, agency=None):
        """Make a request to the NextBus API with the "messages" command to get currently active
        messages to display for multiple routes.

        Arguments:
            route_tags: (List or tuple) List of NextBus route tags for the routes to get messages
                for.
            agency: (String) Name of a transit agency on NextBus. This must be provided if the
                "agency" instance attribute has not been set.

        Returns:
            If the output_format is "json": Dictionary containing the JSON returned by the request.
            If the output_format is "xml": String containing the XML returned by the request.

        Raises:
            urllib.error.HTTPError: If an HTTP error occurs when making the request to the NextBus
                API.
            json.decoder.JSONDecodeError: If the output_format is "json" and the response was not
                valid JSON.
        """

        if not isinstance(route_tags, list) and not isinstance(route_tags, tuple):
            raise TypeError('"route_tags" must be a list or tuple.')

        agency = self._get_agency(agency)

        params = {
            'command': 'messages',
            # Hacky way of repeating the "r" key in the query string for each route
            'r': '&r='.join(route_tags),
            'a': agency
        }

        return self._perform_request(params=params)

    def get_route_list(self, agency=None):
        """Make a request to the NextBus API with the "routeList" command to get a list of routes
        with predictions for a given transit agency.

        Arguments:
            agency: (String) Name of a transit agency on NextBus. This must be provided if the
                "agency" instance attribute has not been set.

        Returns:
            If the output_format is "json": Dictionary containing the JSON returned by the request.
            If the output_format is "xml": String containing the XML returned by the request.

        Raises:
            urllib.error.HTTPError: If an HTTP error occurs when making the request to the NextBus
                API.
            json.decoder.JSONDecodeError: If the output_format is "json" and the response was not
                valid JSON.
        """

        agency = self._get_agency(agency)

        params = {
            'command': 'routeList',
            'a': agency
        }

        return self._perform_request(params=params)

    def get_route_config(self, route_tag=None, agency=None):
        """Make a request to the NextBus API with the "routeConfig" command to get the details of a
        single route, including all stops, and latitude and longitude of the route's path.

        Arguments:
            route_tag: (String) The NextBus route tag for a route. If this is None, the route
                configuration for all routes will be returned.
            agency: (String) Name of a transit agency on NextBus. This must be provided if the
                "agency" instance attribute has not been set.

        Returns:
            If the output_format is "json": Dictionary containing the JSON returned by the request.
            If the output_format is "xml": String containing the XML returned by the request.

        Raises:
            urllib.error.HTTPError: If an HTTP error occurs when making the request to the NextBus
                API.
            json.decoder.JSONDecodeError: If the output_format is "json" and the response was not
                valid JSON.
        """

        agency = self._get_agency(agency)

        params = {
            'command': 'routeConfig',
            'a': agency
        }

        if route_tag is not None:
            params['r'] = route_tag

        return self._perform_request(params=params)

    def get_predictions(self, stop_tag, route_tag, agency=None):
        """Make a request to the NextBus API with the "predictions" command to get arrival time
        predictions for a single stop. A route tag can optionally be provided to filter the
        predictions down to only that particular route at the stop.

        Arguments:
            stop_tag: (String) Tag identifying a stop on NextBus.
            route_tag: (String) The NextBus route tag for a route. If this is None, predictions for
                all routes that serve the given stop will be returned.
            agency: (String) Name of a transit agency on NextBus. This must be provided if the
                "agency" instance attribute has not been set.

        Returns:
            If the output_format is "json": Dictionary containing the JSON returned by the request.
            If the output_format is "xml": String containing the XML returned by the request.

        Raises:
            urllib.error.HTTPError: If an HTTP error occurs when making the request to the NextBus
                API.
            json.decoder.JSONDecodeError: If the output_format is "json" and the response was not
                valid JSON.
        """

        agency = self._get_agency(agency)

        params = {
            'command': 'predictions',
            'a': agency,
            's': stop_tag,
            'r': route_tag,
        }

        return self._perform_request(params=params)

    def get_predictions_for_multi_stops(self, stops, agency=None):
        """Make a request to the NextBus API with the "predictionsForMultiStops" command to get
        arrival time predictions for multiple stops.

        Arguments:
            stops: (List or tuple) List or tuple of dictionaries identifying the combinations of
                routes and stops to get predictions for. Each dictionary must contain the keys
                "stop_tag" and "route_tag", indicating each stop to get predictions for, and the
                route to get predictions for at that stop, respectively.
            agency: (String) Name of a transit agency on NextBus. This must be provided if the
                "agency" instance attribute has not been set.

        Returns:
            If the output_format is "json": Dictionary containing the JSON returned by the request.
            If the output_format is "xml": String containing the XML returned by the request.

        Raises:
            urllib.error.HTTPError: If an HTTP error occurs when making the request to the NextBus
                API.
            json.decoder.JSONDecodeError: If the output_format is "json" and the response was not
                valid JSON.
        """

        if not isinstance(stops, list) and not isinstance(stops, tuple):
            raise TypeError('"stops" must be a list or a tuple.')

        for stop in stops:
            if not isinstance(stop, dict) or 'route_tag' not in stop or 'stop_tag' not in stop:
                raise ValueError('"stops" must contain dictionaries with the "route_tag" and ' \
                                 '"stop_tag" keys')

        agency = self._get_agency(agency)

        params = {
            'command': 'predictionsForMultiStops',
            # Hacky way of repeating the "stops" key in the query string for each route
            'stops': '&stops='.join(['%s|%d' % (stop['route_tag'],
                                                stop['stop_tag']) for stop in stops]),
            'a': agency
        }
        return self._perform_request(params=params)

    def get_schedule(self, route_tag, agency=None):
        """Make a request to the NextBus API with the "schedule" command to get the schedule for a
        single route.

        Arguments:
            route_tag: (String) The NextBus route tag for a route.
            agency: (String) Name of a transit agency on NextBus. This must be provided if the
                "agency" instance attribute has not been set.

        Returns:
            If the output_format is "json": Dictionary containing the JSON returned by the request.
            If the output_format is "xml": String containing the XML returned by the request.

        Raises:
            urllib.error.HTTPError: If an HTTP error occurs when making the request to the NextBus
                API.
            json.decoder.JSONDecodeError: If the output_format is "json" and the response was not
                valid JSON.
        """

        agency = self._get_agency(agency)

        params = {
            'command': 'schedule',
            'a': agency,
            'r': route_tag
        }

        return self._perform_request(params=params)

    def get_vehicle_locations(self, route_tag, timestamp, agency=None):
        """Make a request to the NextBus API with the "vehicleLocations" command to get the all of
        the vehicles on a route with locations that have changed since a given time.

        Arguments:
            route_tag: (String) The NextBus route tag for a route.
            timestamp: (Integer) Unix timestamp in milliseconds indicating that only vehicles with
                locations that have changed since this time will be returned.
            agency: (String) Name of a transit agency on NextBus. This must be provided if the
                "agency" instance attribute has not been set.

        Returns:
            If the output_format is "json": Dictionary containing the JSON returned by the request.
            If the output_format is "xml": String containing the XML returned by the request.

        Raises:
            urllib.error.HTTPError: If an HTTP error occurs when making the request to the NextBus
                API.
            json.decoder.JSONDecodeError: If the output_format is "json" and the response was not
                valid JSON.
        """

        agency = self._get_agency(agency)

        params = {
            'command': 'vehicleLocations',
            'a': agency,
            'r': route_tag,
            't': timestamp
        }

        return self._perform_request(params=params)

    def _get_agency(self, agency):
        """Get the agency name from either the provided argument or "agency" instance attribute.

        Args:
            agency: Name of a transit agency on NextBus.

        Returns:
            The value of the "agency" argument if it is not None, or the value of the instance's
            "agency" atrribute if that is not None.

        Raises:
            ValueError: If both the value of the "agency" argument and this instance's "agency"
                atrribute are None.
        """

        if agency is None:
            if self.agency is None:
                raise ValueError('"agency" is required.')
            else:
                return self.agency
        else:
            return agency

    def _perform_request(self, params):
        """Make a request to the NextBus API with given parameters.

        Arguments:
            params: (Dictionary) Query parameters to provide with the request.

        Returns:
            If the output_format is "json": Dictionary containing the JSON returned by the request.
            If the output_format is "xml": String containing the XML returned by the request.

        Raises:
            urllib.error.HTTPError: If an HTTP error occurs when making the request to the NextBus
                API.
            json.decoder.JSONDecodeError: If the output_format is "json" and the response was not
                valid JSON.
        """

        if self.output_format == 'json':
            base_url = NEXTBUS_JSON_FEED_URL
        elif self.output_format == 'xml':
            base_url = NEXTBUS_XML_FEED_URL
        else:
            raise ValueError('Invalid output_format: %s' % self.output_format)

        url = '%s?%s' % (base_url, urllib.parse.urlencode(params, safe='&='))

        headers = {}

        if self.use_compression:
            headers['Accept-Encoding'] = 'gzip, deflate'

        request = urllib.request.Request(url=url,
                                         headers=headers,
                                         method='GET')

        try:
            LOG.info('Making request to URL %s', url)
            with urllib.request.urlopen(request) as response:
                response_text = response.read()

                if self.output_format == 'json':
                    response = json.loads(response_text)
                elif self.output_format == 'xml':
                    response = response_text

        except urllib.error.HTTPError as exc:
            LOG.error('Request returned status %s due to reason: %s', exc.code, exc.reason)
            raise exc
        except json.decoder.JSONDecodeError as exc:
            LOG.error('Request did not return valid JSON. %s', exc)
            raise exc
        else:
            return response



"""Unit tests for client.py"""

# pylint: disable=unused-argument, protected-access, no-self-use, invalid-name

import unittest
import unittest.mock

import client

class TestInit(unittest.TestCase):
    """Tests for the constructor of the NextBusClient class."""

    def test_type_error_raised_if_output_format_is_not_string(self):
        """Test that a TypeError is raised when the value for the output_format argument is not a
        string."""

        with self.assertRaises(TypeError):
            client.NextBusClient(output_format=None)

        with self.assertRaises(TypeError):
            client.NextBusClient(output_format=1)

        with self.assertRaises(TypeError):
            client.NextBusClient(output_format=True)

        with self.assertRaises(TypeError):
            client.NextBusClient(output_format=['json'])

        with self.assertRaises(TypeError):
            client.NextBusClient(output_format={'output_format': 'json'})

    def test_value_error_raised_if_output_format_is_not_json_or_xml(self):
        """Test that a ValueError is raised if the value for the output_format argument is a string
        but is neither "json" nor "xml"."""

        with self.assertRaises(ValueError):
            client.NextBusClient(output_format='foo')

    def test_no_exceptions_raised_if_output_format_is_valid(self):
        """Test that no exceptions are raised if the value for the output_format argument is either
        "json" or "xml"."""

        client.NextBusClient(output_format='json')
        client.NextBusClient(output_format='JSON')
        client.NextBusClient(output_format='Json')
        client.NextBusClient(output_format='xml')
        client.NextBusClient(output_format='XML')
        client.NextBusClient(output_format='Xml')

    def test_type_error_raised_if_use_compression_is_not_boolean(self):
        """Test that a TypeError is raised if the value for the use_compression argument is not a
        boolean."""

        with self.assertRaises(TypeError):
            client.NextBusClient(output_format='json',
                                 use_compression='True')

        with self.assertRaises(TypeError):
            client.NextBusClient(output_format='json',
                                 use_compression=1)

        with self.assertRaises(TypeError):
            client.NextBusClient(output_format='json',
                                 use_compression=None)

        with self.assertRaises(TypeError):
            client.NextBusClient(output_format='json',
                                 use_compression=[True])

    def test_no_exceptions_raised_if_use_compression_is_boolean(self):
        """Test that no exceptions are raised if the value for the use_compression argument is a
        boolean."""

        client.NextBusClient(output_format='json',
                             use_compression=True)
        client.NextBusClient(output_format='json',
                             use_compression=False)

    def test_attributes_set(self):
        """Test that the instance's attributes are set from the passed arguments."""

        output_format = 'xml'
        agency = 'foo'
        use_compression = False

        nextbus_client = client.NextBusClient(output_format=output_format,
                                              agency=agency,
                                              use_compression=use_compression)

        self.assertEqual(nextbus_client.output_format, output_format)
        self.assertEqual(nextbus_client.agency, agency)
        self.assertEqual(nextbus_client.use_compression, use_compression)

@unittest.mock.patch('client.NextBusClient._perform_request')
class TestGetAgencyList(unittest.TestCase):
    """Tests for the get_agency_list method in the NextBusClient class."""

    def test_request_made_with_agency_list_command(self, perform_request):
        """Test that _perform_request method is called with the "agencyList" command in the
        parameters."""

        nextbus_client = client.NextBusClient(output_format='json')
        nextbus_client.get_agency_list()

        params = perform_request.call_args[1]['params']
        self.assertIn('command', params)
        self.assertEqual(params['command'], 'agencyList')

@unittest.mock.patch('client.NextBusClient._get_agency')
@unittest.mock.patch('client.NextBusClient._perform_request')
class TestGetMessages(unittest.TestCase):
    """Tests for the get_messages method in the NextBusClient class."""

    def test_type_error_raised_if_route_tags_not_list_or_tuple(self, perform_request, get_agency):
        """Test that a TypeError is raised if the value provided for the "route_tags" argument is
        neither a list nor a tuple."""

        nextbus_client = client.NextBusClient(output_format='json')

        with self.assertRaises(TypeError):
            nextbus_client.get_messages(route_tags={})

        with self.assertRaises(TypeError):
            nextbus_client.get_messages(route_tags=None)

        with self.assertRaises(TypeError):
            nextbus_client.get_messages(route_tags='foo')

        with self.assertRaises(TypeError):
            nextbus_client.get_messages(route_tags=1)

        with self.assertRaises(TypeError):
            nextbus_client.get_messages(route_tags=set())

    def test_type_error_not_raised_if_route_tags_is_list_or_tuple(
            self, perform_request, get_agency):
        """Test that a TypeError is not raised if the value provided for the "route_tags" is a tuple
        or a list."""

        nextbus_client = client.NextBusClient(output_format='json')
        nextbus_client.get_messages(route_tags=['foo'],
                                    agency='bar')
        nextbus_client.get_messages(route_tags=('foo',),
                                    agency='bar')

    def test_parameters_passed_to_perform_request(self, perform_request, get_agency):
        """Test that the correct parameters are passed to the _perform_request method."""

        agency = 'foo'
        nextbus_client = client.NextBusClient(output_format='json')

        # Test with one route
        nextbus_client.get_messages(route_tags=['foo'],
                                    agency=agency)
        perform_request.assert_called_once_with(params={
            'command': 'messages',
            'a': get_agency.return_value,
            'r': 'foo'
        })
        get_agency.assert_called_once_with(agency)

        perform_request.reset_mock()

        # Test with multiple routes
        nextbus_client.get_messages(route_tags=['foo', 'bar'],
                                    agency=agency)
        perform_request.assert_called_once_with(params={
            'command': 'messages',
            'a': get_agency.return_value,
            'r': 'foo&r=bar'
        })

@unittest.mock.patch('client.NextBusClient._get_agency')
@unittest.mock.patch('client.NextBusClient._perform_request')
class TestGetRouteList(unittest.TestCase):
    """Tests for the get_route_list method in the NextBusClient class."""

    def test_parameters_passed_to_perform_request(self, perform_request, get_agency):
        """Test that the correct parameters are passed to the _perform_request method."""

        agency = 'foo'
        nextbus_client = client.NextBusClient(output_format='json')
        nextbus_client.get_route_list(agency)

        get_agency.assert_called_once_with(agency)
        perform_request.assert_called_once_with(params={
            'command': 'routeList',
            'a': get_agency.return_value
        })

@unittest.mock.patch('client.NextBusClient._get_agency')
@unittest.mock.patch('client.NextBusClient._perform_request')
class TestGetRouteConfig(unittest.TestCase):
    """Tests for the get_route_config method in the NextBusClient class."""

    def test_parameters_passed_to_perform_request(self, perform_request, get_agency):
        """Test that the correct parameters are passed to the _perform_request method."""

        agency = 'foo'
        nextbus_client = client.NextBusClient(output_format='json')
        nextbus_client.get_route_config(agency=agency)

        get_agency.assert_called_once_with(agency)
        perform_request.assert_called_once_with(params={
            'command': 'routeConfig',
            'a': get_agency.return_value
        })

    def test_route_added_to_parameters_if_route_tag_not_none(self, perform_request, get_agency):
        """Test that the "r" key is added to the parameters passed to the the _perform_request
        method if the value of the route_tag argument is not None."""

        route_tag = 'foo'
        nextbus_client = client.NextBusClient(output_format='json')
        nextbus_client.get_route_config(route_tag=route_tag)

        params = perform_request.call_args[1]['params']
        self.assertIn('r', params)
        self.assertEqual(params['r'], route_tag)

    def test_route_not_added_to_parameters_if_route_tag_is_none(self, perform_request, get_agency):
        """Test that the "r" key is not added to the parameters passed to the _perform_request
        method if the value of the route_tag argument is None."""

        nextbus_client = client.NextBusClient(output_format='json')
        nextbus_client.get_route_config(route_tag=None)

        params = perform_request.call_args[1]['params']
        self.assertNotIn('r', params)

@unittest.mock.patch('client.NextBusClient._get_agency')
@unittest.mock.patch('client.NextBusClient._perform_request')
class TestGetPredictions(unittest.TestCase):
    """Tests for the get_predictions method in the NextBusClient class."""

    def test_parameters_passed_to_perform_request(self, perform_request, get_agency):
        """Test that the correct parameters are passed to the _perform_request method."""

        stop_tag = 12345
        route_tag = 'foo'
        agency = 'foo'
        nextbus_client = client.NextBusClient(output_format='json')
        nextbus_client.get_predictions(stop_tag=stop_tag,
                                       route_tag=route_tag,
                                       agency=agency)

        get_agency.assert_called_once_with(agency)
        perform_request.assert_called_once_with(params={
            'command': 'predictions',
            'a': get_agency.return_value,
            's': stop_tag,
            'r': route_tag,
        })

@unittest.mock.patch('client.NextBusClient._get_agency')
@unittest.mock.patch('client.NextBusClient._perform_request')
class TestGetPredictionsForMultiStops(unittest.TestCase):
    """Tests for the get_predictions_for_multi_stops method in the NextBusClient class."""

    def test_type_error_raised_if_stops_is_not_list_or_tuple(self, perform_request, get_agency):
        """Test that a TypeError is raised if the value for the "stops" argument is neither a list
        nor a tuple."""

        nextbus_client = client.NextBusClient(output_format='json')

        with self.assertRaises(TypeError):
            nextbus_client.get_predictions_for_multi_stops(stops={})

        with self.assertRaises(TypeError):
            nextbus_client.get_predictions_for_multi_stops(stops=set())

        with self.assertRaises(TypeError):
            nextbus_client.get_predictions_for_multi_stops(stops=None)

        with self.assertRaises(TypeError):
            nextbus_client.get_predictions_for_multi_stops(stops=True)

        with self.assertRaises(TypeError):
            nextbus_client.get_predictions_for_multi_stops(stops='foo')

        with self.assertRaises(TypeError):
            nextbus_client.get_predictions_for_multi_stops(stops=1)

    def test_value_error_raised_if_stops_does_not_contain_dictionaries_with_required_keys(
            self, perform_request, get_agency):
        """Test that a ValueError is raised if the value for the "stops" argument is a list or a
        tuple but does not contain dictionaries with the required keys."""

        nextbus_client = client.NextBusClient(output_format='json')

        with self.assertRaises(ValueError):
            nextbus_client.get_predictions_for_multi_stops(stops=['foo'])

        with self.assertRaises(ValueError):
            nextbus_client.get_predictions_for_multi_stops(stops=[1])

        with self.assertRaises(ValueError):
            nextbus_client.get_predictions_for_multi_stops(stops=[['foo']])

        with self.assertRaises(ValueError):
            nextbus_client.get_predictions_for_multi_stops(stops=[{}])

        with self.assertRaises(ValueError):
            nextbus_client.get_predictions_for_multi_stops(stops=[{
                'stop_tag': 1234
            }])
        with self.assertRaises(ValueError):
            nextbus_client.get_predictions_for_multi_stops(stops=[{
                'route_tag': 'foo'
            }])
        with self.assertRaises(ValueError):
            nextbus_client.get_predictions_for_multi_stops(
                stops=[
                    {
                        'stop_tag': 1234,
                        'route_tag': 'foo'
                    },
                    None
                ]
            )
        with self.assertRaises(ValueError):
            nextbus_client.get_predictions_for_multi_stops(stops=('foo',))

        with self.assertRaises(ValueError):
            nextbus_client.get_predictions_for_multi_stops(stops=(1,))

        with self.assertRaises(ValueError):
            nextbus_client.get_predictions_for_multi_stops(stops=(('foo'),))

        with self.assertRaises(ValueError):
            nextbus_client.get_predictions_for_multi_stops(stops=({},))

        with self.assertRaises(ValueError):
            nextbus_client.get_predictions_for_multi_stops(stops=({
                'stop_tag': 1234
            },))
        with self.assertRaises(ValueError):
            nextbus_client.get_predictions_for_multi_stops(stops=({
                'route_tag': 'foo'
            },))
        with self.assertRaises(ValueError):
            nextbus_client.get_predictions_for_multi_stops(
                stops=(
                    {
                        'stop_tag': 1234,
                        'route_tag': 'foo'
                    },
                    None
                )
            )

    def test_no_exceptions_raised_if_stops_is_valid(self, perform_request, get_agency):
        """Test that no exceptions are raised if the value for the "stops" argument is either a list
        or a tuple containing dictionaries with the required keys."""

        nextbus_client = client.NextBusClient(output_format='json')

        nextbus_client.get_predictions_for_multi_stops(stops=[{
            'route_tag': 'foo',
            'stop_tag': 1234
        }])

        nextbus_client.get_predictions_for_multi_stops(stops=({
            'route_tag': 'foo',
            'stop_tag': 1234
        },))

    def test_parameters_passed_to_perform_request(self, perform_request, get_agency):
        """Test that the correct parameters are passed to the _perform_request method."""

        agency = 'foo'
        nextbus_client = client.NextBusClient(output_format='json')

        # Test with only one stop
        stop_tag = 1234
        route_tag = 'bar'
        nextbus_client.get_predictions_for_multi_stops(
            agency=agency,
            stops=[{
                'route_tag': route_tag,
                'stop_tag': stop_tag
            }]
        )

        get_agency.assert_called_once_with(agency)
        perform_request.assert_called_once_with(params={
            'command': 'predictionsForMultiStops',
            'a': get_agency.return_value,
            'stops': '%s|%s' % (route_tag, stop_tag)
        })

        perform_request.reset_mock()

        # Test with multiple stops
        first_stop_tag = 1234
        first_route_tag = 'baz'
        second_stop_tag = 5678
        second_route_tag = 'buz'
        nextbus_client.get_predictions_for_multi_stops(
            agency=agency,
            stops=[
                {
                    'route_tag': first_route_tag,
                    'stop_tag': first_stop_tag
                },
                {
                    'route_tag': second_route_tag,
                    'stop_tag': second_stop_tag
                }
            ]
        )

        perform_request.assert_called_once_with(params={
            'command': 'predictionsForMultiStops',
            'a': get_agency.return_value,
            'stops': '%s|%s&stops=%s|%s' % (first_route_tag, first_stop_tag, second_route_tag,
                                            second_stop_tag)
        })

@unittest.mock.patch('client.NextBusClient._get_agency')
@unittest.mock.patch('client.NextBusClient._perform_request')
class TestGetSchedule(unittest.TestCase):
    """Tests for the get_schedule method in the NextBusClient class."""

    def test_parameters_passed_to_perform_request(self, perform_request, get_agency):
        """Test that the correct parameters are passed to the _perform_request method."""

        route_tag = 'foo'
        nextbus_client = client.NextBusClient(output_format='json')
        nextbus_client.get_schedule(route_tag=route_tag)

        perform_request.assert_called_once_with(params={
            'command': 'schedule',
            'a': get_agency.return_value,
            'r': route_tag
        })

@unittest.mock.patch('client.NextBusClient._get_agency')
@unittest.mock.patch('client.NextBusClient._perform_request')
class TestGetVehicleLocations(unittest.TestCase):
    """Tests for the get_vehicle_locations method in the NextBusClient class."""

    def test_parameters_passed_to_perform_request(self, perform_request, get_agency):
        """Test that the correct parameters are passed to the _perform_request method."""

        route_tag = 'foo'
        timestamp = 12345
        nextbus_client = client.NextBusClient(output_format='json')
        nextbus_client.get_vehicle_locations(route_tag=route_tag,
                                             timestamp=timestamp)

        perform_request.assert_called_once_with(params={
            'command': 'vehicleLocations',
            'a': get_agency.return_value,
            'r': route_tag,
            't': timestamp
        })

class TestGetAgency(unittest.TestCase):
    """Tests for the _get_agency method in the NextBusClient class."""

    def test_agency_argument_returned_if_not_none(self):
        """Test that the value of the "agency" argument is returned if it is not None."""

        nextbus_client = client.NextBusClient(output_format='json')

        agency = 'foo'
        nextbus_client.agency = 'bar'

        response = nextbus_client._get_agency(agency)

        assert response == agency

    def test_agency_attribute_returned_if_not_none_and_agency_argument_is_none(self):
        """Test that the value of the NextBusClient instance's "agency" atrribute is returned if it
        is not None and the value of the "agency" argument is None."""

        nextbus_client = client.NextBusClient(output_format='json')
        nextbus_client.agency = 'foo'

        response = nextbus_client._get_agency(None)

        assert response == nextbus_client.agency

    def test_value_error_raised_if_agency_argument_and_agency_attribute_are_none(self):
        """Test that a ValueError is raised if both the "agency" argument and the NextBusClient
        instance's "agency" attribute are None."""

        nextbus_client = client.NextBusClient(output_format='json')
        nextbus_client.agency = None

        with self.assertRaises(ValueError):
            nextbus_client._get_agency(None)

@unittest.mock.patch('client.json.loads')
@unittest.mock.patch('client.urllib.request')
class TestPerformRequest(unittest.TestCase):
    """Tests for the _perform_request method in the NextBusClient class."""

    def test_json_output_format_makes_request_to_nextbus_json_feed(self, request, loads):
        """Test that requests made to the NextBus API when the instance's output_format is "json"
        are made to the NextBus public JSON feed."""

        nextbus_client = client.NextBusClient(output_format='json')
        nextbus_client._perform_request(params={})

        # Get the URL of the request to the NextBus API
        request_url = request.Request.call_args[1]['url']
        assert request_url.startswith(client.NEXTBUS_JSON_FEED_URL)
        request.urlopen.assert_called_once_with(request.Request.return_value)

    def test_xml_output_format_makes_request_to_nextbus_xml_feed(self, request, loads):
        """Test that requests made to the NextBus API when the instance's output_format is "xml"
        are made to the NextBus public XML feed."""

        nextbus_client = client.NextBusClient(output_format='xml')
        nextbus_client._perform_request(params={})

        # Get the URL of the request to the NextBus API
        request_url = request.Request.call_args[1]['url']
        assert request_url.startswith(client.NEXTBUS_XML_FEED_URL)
        request.urlopen.assert_called_once_with(request.Request.return_value)

    def test_value_error_raised_if_output_format_is_not_json_or_xml(self, request, loads):
        """Test that a ValueError is raised when the instance's output_format is neither "xml" nor
        "json"."""

        nextbus_client = client.NextBusClient(output_format='json')
        nextbus_client.output_format = 'foo'

        with self.assertRaises(ValueError):
            nextbus_client._perform_request(params={})

    @unittest.mock.patch('client.urllib.parse.urlencode')
    def test_parameters_added_to_url_as_query_string(self, urlencode, request, loads):
        """Test that the provided parameters are converted to a query string and added to the URL
        used for making a request to the NextBus API."""

        params = unittest.mock.MagicMock()
        nextbus_client = client.NextBusClient(output_format='json')
        nextbus_client._perform_request(params=params)

        urlencode.assert_called_once_with(params, safe='&=')
        # Get the URL of the request to the NextBus API
        request_url = request.Request.call_args[1]['url']
        assert request_url.endswith('?%s' % urlencode.return_value)
        request.urlopen.assert_called_once_with(request.Request.return_value)

    def test_accept_encoding_header_in_request_if_use_compression_is_true(self, request, loads):
        """Test that the Accept-Encoding header is present in the request to the NextBus API if the
        use_compression attribute is True."""

        nextbus_client = client.NextBusClient(output_format='json',
                                              use_compression=True)
        nextbus_client._perform_request(params={})

        headers = request.Request.call_args[1]['headers']
        self.assertIn('Accept-Encoding', headers)
        self.assertEqual(headers['Accept-Encoding'], 'gzip, deflate')

    def test_accept_encoding_header_not_in_request_if_use_compression_is_false(
            self, request, loads):
        """Test that the Accept-Encoding header is not present in the request to the NextBus API if
        the use_compression attribute is False."""

        nextbus_client = client.NextBusClient(output_format='json',
                                              use_compression=False)
        nextbus_client._perform_request(params={})

        headers = request.Request.call_args[1]['headers']
        self.assertNotIn('Accept-Encoding', headers)

    def test_response_returned_as_dictionary_if_output_format_is_json(self, request, loads):
        """Test that the response from the NextBus API is loaded as JSON and returned as a
        dictionary if the output_format is "json"."""

        nextbus_client = client.NextBusClient(output_format='json')
        response = nextbus_client._perform_request(params={})

        loads.assert_called_once_with(
            request.urlopen.return_value.__enter__.return_value.read.return_value
        )
        self.assertEqual(response, loads.return_value)

    def test_response_returned_as_string_if_output_format_is_xml(self, request, loads):
        """Test that the response from the NextBus API is directly returned if the output_format is
        "xml"."""

        nextbus_client = client.NextBusClient(output_format='xml')
        response = nextbus_client._perform_request(params={})

        self.assertEqual(response,
                         request.urlopen.return_value.__enter__.return_value.read.return_value)

    @unittest.mock.patch('client.LOG')
    def test_error_logged_if_http_error_raised(self, log, request, loads):
        """Test that an error is logged if a urllib HTTPError is raised when making the request to
        the NextBus API."""

        request.urlopen.side_effect = client.urllib.error.HTTPError(url='foo',
                                                                    code=500,
                                                                    msg='bar',
                                                                    hdrs={},
                                                                    fp=unittest.mock.MagicMock())

        nextbus_client = client.NextBusClient(output_format='json')
        with self.assertRaises(client.urllib.error.HTTPError):
            nextbus_client._perform_request(params={})

        log.error.assert_called_once()

    @unittest.mock.patch('client.LOG')
    def test_error_logged_if_json_decode_error_raised(self, log, request, loads):
        """Test that an error is logged if a JSONDecodeError is raised when making the request to
        the NextBus API."""

        request.urlopen.side_effect = client.json.decoder.JSONDecodeError(msg='foo',
                                                                          doc='bar',
                                                                          pos=0)

        nextbus_client = client.NextBusClient(output_format='json')
        with self.assertRaises(client.json.decoder.JSONDecodeError):
            nextbus_client._perform_request(params={})

        log.error.assert_called_once()

if __name__ == '__main__':
    unittest.main()



from setuptools import setup, find_packages

with open('README.md', 'r') as readme:
    long_description = readme.read()

setup(
    name='py_nextbusnext',
    version='0.1.3',
    author='ViViDboarder',
    description='Minimalistic Python client for the NextBus public API for real-time transit ' \
                'arrival data',
    long_description=long_description,
    long_description_content_type="text/markdown",
    test_suite='tests.py',
    url='https://github.com/vividboarder/py_nextbus',
    packages=find_packages(),
    python_requires='>=3',
    classifiers=(
        "Programming Language :: Python :: 3.6 ",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ),
)
